#ifndef _MENSAGEM_H_
#define _MENSAGEM_H_
#include "mpi.h"
#include "constantes.h"
#include "buffer.h"
#include "fifo.h"

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


extern MPI_Datatype tipo[12];

typedef struct STATUS {
  int source;
  int tag;
  int size;
} mensagem_status;

typedef struct mensagem_fila {
  int source;
  int tag;
  int size;
  buffer *buf;
} Mensagem_fila;

class mensagem {

 protected:

  /* Rank do monitor */

  int rank;

  /* Relogio logico das mensagens */

  int logical_clock;

  /* Contador de ckeckpoints (incrementa a cada checkpoint - 
     cada algoritmo deve ter a sua propria forma de controlar
     quando um ckeckpoint forcado deve ser tirado*/
  
  int checkpoint_count;
  
  /*Flag que indica se o envio / recebimento de mensagens est bloqueado 
    ou no*/

  int blocked;

  /*Fila de mensagens recebidas */

  fifo *fila;

  /*Funcoes de envio de mensagens - baixo nivel
    Codifica o envio das mensagens utilizando a MPI desejada (LAM-MPI, TCP
    ... )*/

  int send (buffer *buf, int dest, int tag);
  int receive (buffer **buf, int dest, int tag, mensagem_status *status);
  int take_checkpoint(int forced);

  /*Funcoes virtuais que devem ser sobrescritas pelos filhos para 
    determinar o funcionamento de cada protocolo*/

  virtual void informacoes_controle(buffer *b);
  virtual int processa_recebimento(buffer *b, mensagem_status *status);
  virtual int verifica_mensagem_controle (int tag);

 public:
  
  mensagem(int *argc, char** argv[], void (*grava_estado)(buffer *estado));
  virtual ~mensagem();
  int monitor ();
  int tamanho_rede();
  int id_processo();
  int timestamp();
  int monitora(buffer **buf, int dest, int tag, mensagem_status *status);
  virtual void tira_checkpoint_basico();
  virtual int envia_mensagem(buffer *buf, int dest, int tag);
  int recebe_mensagem(buffer **buf, int source, int tag, mensagem_status *status, int tipo);
  int recebe_mensagem(buffer **buf, int source, int tag, mensagem_status *status);
  int recebe_mensagem(int tipo);
  virtual char* algoritmo_utilizado();

  /*Funcoes de inicializacao / encerramento de processos*/
  virtual void inicializa_monitor();
  virtual void inicializa_monitor(buffer ***retorno);
  virtual void inicializa_processo();
  virtual void aguarda_desbloqueio();
  virtual void finaliza_processo();

  /*Ponteiro para funcao de gravacao de estado*/
  void (*coleta_estado) (buffer *estado);
};
#endif
