#include "mensagem.h"
#include "stdlib.h"
#include "bcs.h"
#include "iostream.h"
#include "mpi.h"

#define REQUEST_TAG 1
#define PARENT_TAG 2

#define PAI_TAG 3
#define ARVORE_TAG 4
#define ORDEM_TAG 5

#define NIL -1

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


int pai = NIL;

int * arvore;
int * ordem;
int tamanho_arvore;

void carrega_estado (buffer *estado) {
  /*Informando o pai*/
  estado->adiciona_pacote((void *)&pai, MSG_INT, 1, PAI_TAG);

  /*Informando a arvore*/
  estado->adiciona_pacote((void *)arvore, MSG_INT, tamanho_arvore, ARVORE_TAG);

  /*Informando a arvore*/
  estado->adiciona_pacote((void *)ordem, MSG_INT, tamanho_arvore, ORDEM_TAG);

}

/*Algoritmo que encontra uma arvore geradora a partir de uma lista local de conexoes, sem pesos
  Para garantir que o grafo seja ao menos conexo, todo vertice i ter uma aresta at o vertice 
  i + 1, e outras arestas geradas aleatoriamente. O envio das mensagens ocorrer na ordem de rank
  Somente um processo comeara a gerao da arvore. Assim que todos os filhos forem notificados, a
  execuo termina*/

int main (int argc, char* argv[]) {
  
  mensagem* M;
  buffer *B;
  mensagem_status s;
  int proximo_processo;
  void *vetor;
  int tag_p;
  int tipo_p;
  int tamanho_p;
  int i;

  M = new bcs (&argc, &argv, carrega_estado);

  B = NULL;

  srand(M->id_processo());

  tamanho_arvore = M->tamanho_rede();
  arvore = (int *) calloc (tamanho_arvore, sizeof(int));
  ordem = (int *) calloc (tamanho_arvore, sizeof(int));
  for (i=0; i < tamanho_arvore; i++) {
    arvore[i] = 0; /*Nao pertence a arvore*/
    ordem[i] = -1; /*Nenhum processo ocupa essa posio*/
  }

  /*Utiliza o vetor arvore como auxiliar para gerar uma ordem aleatria
    de pesquisa na arvore - serve para randomizar a arvore geradora*/
  
  

  M->inicializa_processo();

  /*Define o prximo processo - ele obrigatoriamente est alocado 
    na arvore*/


  proximo_processo = ((M->id_processo() + 1) % M->tamanho_rede());
  if (proximo_processo == M->monitor()) { 
      proximo_processo = ((proximo_processo + 1) % M->tamanho_rede());
  }

  arvore[proximo_processo] = 1; /*Esse processo est na arvore*/

  M->aguarda_desbloqueio();

  M->finaliza_processo();

  return (0);
}
