#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "bhmr.h"
#include "mpi.h"

#define BHMR_DV_TAG 30000
#define BHMR_SIMPLE_TAG 30001
#define BHMR_SENT_TAG 30002
#define BHMR_CAUSAL_TAG 30003
#define BHMR_CAUSAL_SIZE_TAG 30004

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo BHMR
  Criada por Raphael Marcos Menderico
  Criacao: 01/12/2003
  Ultima alteracao: 02/12/2003
*/

bhmr::bhmr(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  /*A rede j foi inicializada*/
  int i, j;
  /*Alocando vetores*/
  this->dv_size = this->tamanho_rede();
  this->dv = (int *) calloc (this->dv_size, sizeof(int));
  this->simple = (int *) calloc (this->dv_size, sizeof(int));
  this->sent = (int *) calloc (this->dv_size, sizeof(int));
  this->dv_pid = this->id_processo();
  this->causal = (int **) calloc (this->dv_size, sizeof(int *));
  for (i=0; i< this->dv_size; i++) {
    this->causal[i] = (int *) calloc (this->dv_size, sizeof(int));
  }
  /*Inicializando vetores*/
  for (i = 0; i < this->dv_size; i++) {
    for (j = 0; j < this->dv_size; j++) {
      this->causal [i][j] = 0; /*False*/
    }
    this->causal[i][i] = 1; /*True*/
    this->sent[i] = 0; /*False*/
    this->simple[i] = 0; /*False*/
    this->dv[i] = 0;
  }
  this->simple[this->dv_pid] = 1; /*true*/   
  this->dv[this->dv_pid] ++;
}

bhmr::~bhmr() {
  int i;
  free(this->dv);
  free(this->simple);
  free(this->sent);
  for (i = 0; i < this->dv_pid; i++) {
    free(this->causal[i]);
  }
  free(this->causal);
}

int bhmr::processa_recebimento(buffer *b, mensagem_status *status) {

  void *vetor;
  int tipo_p, tamanho_p, tag_p;

  int *dv_msg;
  int *simple_msg;
  int **causal_msg;
  int causal_size_msg;
  int i, j;
  
  int c1;
  int c2;

  /*Alocacao de memria*/
  b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
  causal_size_msg = ((int) *((int*) vetor));

  causal_msg = (int **) calloc (causal_size_msg, sizeof(int *));

  for (i = 0; i < causal_size_msg; i++) {
    b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
    causal_msg[i] = ((int *) vetor);
  }

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  simple_msg = ((int *) vetor);

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  dv_msg = ((int *) vetor);

  /*Algoritmo de checkpoint forcado*/
  c1 = (this->dv[this->dv_pid] == dv_msg[this->dv_pid]) && 
    (!simple_msg[this->dv_pid]);
  c2 = 0; /*False*/ 
  for (i = 0; i < this->dv_size; i++) {
    if ((this->sent[i]) && (i != this->monitor())) {
      for (j = 0; j < this->dv_size; j++) {
	c2 = c2 || ((dv_msg[j] > this->dv[j]) && (!causal_msg[j][i]));
      }
    }
  }
  if (c1 || c2) {
    this->dv[this->dv_pid]++;
    for (i=0; i<this->dv_size; i++) {
      this->sent[i] = 0; /*False*/
      if (i != this->dv_pid) {
	this->simple[i] = 0; /*False*/
	this->causal[this->dv_pid][i] = 0;/*False*/
      }
    }
    this->take_checkpoint(1);
  }
  for (i = 0; i < this->dv_size; i++) {
    if (dv_msg[i] > this->dv[i]) {
      this->dv[i] = dv_msg[i];
      this->simple[i] = simple_msg[i];
      for (j = 0; j < this->dv_size; j++) {
	this->causal[i][j] = causal_msg[i][j];
      }
    } else if (dv_msg[i] == this->dv[i]) {
      this->simple[i] = (this->simple[i] && simple_msg[i]);
      for (j = 0; j < this->dv_size; j++) {
	this->causal[i][j] = (this->causal[i][j] || causal_msg[i][j]);
      }
    }
  }
  
  this->causal[status->source][this->dv_pid] = 1; /*True*/
  for (i = 0; i < this->dv_size; i++) {
    this->causal[i][this->dv_pid] = this->causal[i][this->dv_pid] || 
      this->causal[i][status->source];
  }

  /*Liberacao de memria*/
  for(i=0; i < causal_size_msg; i++) {
    free(causal_msg[i]);
  }

  free(causal_msg);
  free(simple_msg);
  free(dv_msg); 

  return(0);

}

void bhmr::informacoes_controle(buffer *b) {
  /*Ao solicitar as informacoes de controle, isso significa que 
    um checkpoint esta sendo tirado pelo sistema*/
  b->adiciona_pacote((void *)(this->dv), MSG_INT, this->dv_size, BHMR_DV_TAG);
}

void bhmr::tira_checkpoint_basico() {
  int i;
  this->dv[this->dv_pid]++;
  for (i = 0; i < this->dv_size; i++) {
    this->sent[i] = 0; /*False*/
    if (i != this->dv_pid) {
      this->simple[i] = 0; /*False*/
      this->causal[this->dv_pid][i] = 0; /*False*/
    }
  }
  mensagem::tira_checkpoint_basico();
}

int bhmr::envia_mensagem(buffer *buf, int dest, int tag) {
  /*Adicionando o relogio a mensagem enviada*/
  
  int i;

  this->sent[dest] = 1; /*True*/

  buf->adiciona_pacote((void *)(this->dv), MSG_INT, 
		       this->dv_size, BHMR_DV_TAG);

  buf->adiciona_pacote((void *)(this->simple), MSG_INT, 
		       this->dv_size, BHMR_SIMPLE_TAG);
  
  for (i = 0; i< this->dv_size; i++) {
    buf->adiciona_pacote((void *)(this->causal[i]), MSG_INT, 
			 this->dv_size, BHMR_CAUSAL_TAG);
  }

  buf->adiciona_pacote((void *)&(this->dv_size), MSG_INT, 
		       1, BHMR_CAUSAL_SIZE_TAG);

  mensagem::envia_mensagem(buf, dest, tag);
  return(0);
}

char* bhmr::algoritmo_utilizado() {
  return ("BHMR\n");
}
void bhmr::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void bhmr::inicializa_processo() {
  mensagem::inicializa_processo();
}

