#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "bqc.h"
#include "mpi.h"

#define BQC_DV_TAG 30000
#define BQC_PRED_TAG 30001
#define BQC_PRED_SIZE_TAG 30002

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo BQC
  Criada por Raphael Marcos Menderico
  Criacao: 09/12/2003
  Ultima alteracao: 09/12/2003
*/

bqc::bqc(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  /*A rede j foi inicializada*/
  int i, j;
  /*Alocando vetores*/

  this->dv_size = this->tamanho_rede();
  this->dv = (int *) calloc (this->dv_size, sizeof(int));
  this->ipred = (int *) calloc (this->dv_size, sizeof(int));
  this->pred = (int **) calloc (this->dv_size, sizeof(int *));
  for (i=0; i< this->dv_size; i++) {
    this->pred[i] = (int *) calloc (this->dv_size, sizeof(int));
  }
  this->dv_pid = this->id_processo();

  /*Inicializando vetores*/
  this->sent = 0; /*False*/
  for (i = 0; i < this->dv_size; i++) {
    for (j = 0; j < this->dv_size; j++) {
      this->pred [i][j] = -1; 
    }
    this->ipred[i] = -1; 
    this->dv[i] = 0;
  }
  this->dv[this->dv_pid]++;
}

bqc::~bqc() {
  int i;
  free(this->dv);
  free(this->ipred);
  for (i = 0; i < this->dv_pid; i++) {
    free(this->pred[i]);
  }
  free(this->pred);
}

int bqc::processa_recebimento(buffer *b, mensagem_status *status) {

  void *vetor;
  int tipo_p, tamanho_p, tag_p;
  int *dv_msg;
  int **pred_msg;
  int pred_size_msg;
  int i, j;
  
  int teste, valor_max;

  /*Alocacao de memria*/
  b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
  pred_size_msg = ((int) *((int*) vetor));

  pred_msg = (int **) calloc (pred_size_msg, sizeof(int *));

  for (i = 0; i < pred_size_msg; i++) {
    b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
    pred_msg[i] = ((int *) vetor);
  }

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  dv_msg = ((int *) vetor);

  /*Algoritmo de checkpoint forcado*/

  teste = 0; /*False*/ 
  if (this->sent) { 
    /*Parece estranho esse teste, mas ele foi feito para otimizar
      o algoritmo, pois se sent = false, nao ha necessidade de 
      realizar o segundo teste, que eh custoso*/
    for (i = 0; i < this->dv_size; i++) {
      if ((dv_msg[i] > this->dv[i]) && (i != this->monitor())) {
	for (j = 0; j < this->dv_size; j++) {
	  if (dv_msg[j] > this->dv[j]) {
	    valor_max = dv_msg[j];
	  } else {
	    valor_max = this->dv[j];
	  }
	  teste = teste || ((pred_msg[i][j]+1) > valor_max);
	}
      }
    }
    if (teste) {

      this->dv[this->dv_pid]++;
      this->sent = 0; /*False*/
      for (i = 0; i < this->dv_size; i++) {
	if (this->ipred[i] > this->pred[this->dv_pid][i]) {
	  this->pred[this->dv_pid][i] = this->ipred[i]; 
	}
	this->ipred[i] = -1; 
      }
      this->take_checkpoint(1);
    }
  }
  
  for (i=0; i < this->dv_size; i++) {
    if (dv_msg[i] > this->dv[i]) {
      this->dv[i] = dv_msg[i];
    }
    for (j=0; j < this->dv_size; j++) {
      if (pred_msg[i][j] > this->pred[i][j]) {
	this->pred[i][j] = pred_msg[i][j];
      }
    }
  }
  
  if (dv_msg[status->source] > this->ipred[status->source]) {
    this->ipred[status->source] = dv_msg[status->source];
  }
  
  /*Liberacao de memria*/
  for(i=0; i < pred_size_msg; i++) {
    free(pred_msg[i]);
  }

  free(pred_msg);
  free(dv_msg); 

  return(0);

}

void bqc::informacoes_controle(buffer *b) {
  /*Ao solicitar as informacoes de controle, isso significa que 
    um checkpoint esta sendo tirado pelo sistema*/
  b->adiciona_pacote((void *)(this->dv), MSG_INT, this->dv_size, BQC_DV_TAG);
}

void bqc::tira_checkpoint_basico() {
  int i;
  this->dv[this->monitor()] = -1;
  this->dv[this->dv_pid]++;
  this->sent = 0; /*False*/
  for (i = 0; i < this->dv_size; i++) {
    if (this->ipred[i] > this->pred[this->dv_pid][i]) {
      this->pred[this->dv_pid][i] = this->ipred[i]; 
    }
    this->ipred[i] = -1; 
  }
  mensagem::tira_checkpoint_basico();
}

int bqc::envia_mensagem(buffer *buf, int dest, int tag) {

  /*Adicionando o relogio a mensagem enviada*/
  
  int i;

  this->sent = 1; /*True*/

  buf->adiciona_pacote((void *)(this->dv), MSG_INT, 
		       this->dv_size, BQC_DV_TAG);

  for (i = 0; i< this->dv_size; i++) {
    buf->adiciona_pacote((void *)(this->pred[i]), MSG_INT, 
			 this->dv_size, BQC_PRED_TAG);
  }

  buf->adiciona_pacote((void *)&(this->dv_size), MSG_INT, 
		       1, BQC_PRED_SIZE_TAG);

  mensagem::envia_mensagem(buf, dest, tag);
  return(0);
}

char* bqc::algoritmo_utilizado() {
  return ("BQC\n");
}
void bqc::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void bqc::inicializa_processo() {
  mensagem::inicializa_processo();
}

