#include <buffer.h>

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

MPI_Datatype tipo[12] = {
    MPI_CHAR,
    MPI_SHORT,
    MPI_INT,
    MPI_LONG,
    MPI_UNSIGNED_CHAR,
    MPI_UNSIGNED_SHORT,
    MPI_UNSIGNED,
    MPI_UNSIGNED_LONG,
    MPI_FLOAT,
    MPI_DOUBLE,
    MPI_LONG_DOUBLE,
    MPI_BYTE
};

buffer::buffer() {
  /*Aloca o no cabeca*/
  this->dados = (Lista *) calloc (1, sizeof(Lista));
  this->dados->ant = this->dados;
  this->dados->prox = this->dados;
  this->dados->vetor = NULL;
  this->dados->tipo_entrada = 0;
  this->dados->tamanho_entrada = 0;
  this->dados->tag = 0;
  /*Inicializa o contador*/
  this->numero = 0;
}

buffer::~buffer() {
  /*Desaloca todos os registros existentes*/
  int i, tamanho;
  void *vetor;
  tamanho = this->tamanho_atual();
  for (i=0; i<tamanho; i++) {
    this->remove_pacote(&vetor, NULL, NULL, NULL);
    /*    if (vetor != NULL) {
      free(vetor);
      }*/
  }
  /*Desaloca o no cabeca*/
  if (this->dados != NULL) {
    free(this->dados);
  }
}

int buffer::adiciona_pacote(void *vetor, int tipo_entrada, int tamanho_entrada, int tag) {
  Lista *Pacote;
  /*Aloca o novo no e inicializa os valores*/  
  Pacote = (Lista *) calloc (1, sizeof(Lista));
  Pacote->vetor = vetor;
  Pacote->tipo_entrada = tipo_entrada;
  Pacote->tamanho_entrada = tamanho_entrada;
  Pacote->tag = tag;
  /*Adiciona nó no topo da lista ligada (como se fosse uma pilha)*/
  Pacote->ant = this->dados;
  Pacote->prox = this->dados->prox;
  this->dados->prox->ant = Pacote;
  this->dados->prox = Pacote;
  this->numero++;
  return (1);
}

int buffer::remove_pacote(void **vetor, int *tipo_entrada, int *tamanho_entrada,int *tag){
  /*Remove o no atual do topo da lista ligada*/
  Lista *Pacote;
  if (this->numero > 0) {
    Pacote = this->dados->prox; 
    this->dados->prox = Pacote->prox;
    this->dados->prox->ant = this->dados;
    
    /*Retornando as informacoes necessarias*/
    if (vetor != NULL) {
      (*vetor) = Pacote->vetor;
    }
    if (tipo_entrada != NULL) {
      (*tipo_entrada) = Pacote->tipo_entrada;
    }
    if (tamanho_entrada != NULL) { 
      (*tamanho_entrada) = Pacote->tamanho_entrada;
    }
    
    if (tag != NULL) {
      (*tag) = Pacote->tag;
    }
    
    free(Pacote);
    
    this->numero--;
    return (1);
  } else {
    return (0);
  }
}

void buffer::empacota(void **buffer, int *size) {
  int elementos, i, tamanho_total = 0, tamanho_parcial, position;
  void *buf;
  Lista *Pacote;

  /*Calcula o tamanho de cada componente para montar o pacote*/
  elementos = this->tamanho_atual();

  Pacote = this->dados->ant; /*O empacotamento acontece ao contrario, para 
			       facilitar o desempacotamento*/
  /*Calcula o tamanho do contador*/
  MPI_Pack_size(1, tipo[MSG_INT], MPI_COMM_WORLD, &tamanho_parcial);
  tamanho_total = tamanho_parcial;

  for (i=0; i < elementos; i++) {
    /*Calculando o tamanho para enviar os tres inteiros de controle 
      (tipo, tamanho e tag)*/
    MPI_Pack_size(3, tipo[MSG_INT], MPI_COMM_WORLD, &tamanho_parcial);
    tamanho_total += tamanho_parcial;
    /*Calculando o tamanho para enviar o vetor*/
    MPI_Pack_size(Pacote->tamanho_entrada, tipo[Pacote->tipo_entrada], MPI_COMM_WORLD, &tamanho_parcial);
    tamanho_total += tamanho_parcial;

    Pacote = Pacote->ant;
  }
  
  /*Aloca buffer temporario*/
  buf = malloc (tamanho_total);
  /*Cria pacote para ser enviado, retornando o buffer e seu tamanho*/
  position = 0;
  Pacote = this->dados->ant;

  MPI_Pack((void *)&elementos, 1, tipo[MSG_INT], buf, tamanho_total, &position, MPI_COMM_WORLD);

  for (i=0; i < elementos; i++) { 

    MPI_Pack((void *)&(Pacote->tipo_entrada), 1, tipo[MSG_INT], buf, tamanho_total, &position, MPI_COMM_WORLD);

    MPI_Pack((void *)&(Pacote->tamanho_entrada), 1, tipo[MSG_INT], buf, tamanho_total, &position, MPI_COMM_WORLD);

    MPI_Pack((void *)&(Pacote->tag), 1, tipo[MSG_INT], buf, tamanho_total, &position, MPI_COMM_WORLD);

    MPI_Pack(Pacote->vetor, Pacote->tamanho_entrada, tipo[Pacote->tipo_entrada], buf, tamanho_total, &position, MPI_COMM_WORLD);
    Pacote = Pacote->ant;
  }
  (*buffer) = buf;
  (*size) = position;
}

int buffer::tamanho_atual() {
  return (this->numero);
}

int buffer::tag_topo() {
  if (this->numero > 0) { 
    return(this->dados->prox->tag);
  } else {
    return (0);
  }
}

void buffer::desempacota(void *buffer, int size) {
  /*Desempacota o pacote recebido, criando a pilha para ser
    removida posteriormente*/
  int position=0;
  int numelementos;
  int tamanho_buffer;
  int tipo_entrada, tamanho_entrada, tag;
  void *vetor;
  int i;

  /*Desempacota o numero de pacotes*/
  MPI_Unpack (buffer, size, &position, (void *) &numelementos, 1, tipo[MSG_INT], MPI_COMM_WORLD);
  for (i=0; i<numelementos; i++) {
    /*Desempacota as informacoes de controle*/
    MPI_Unpack(buffer, size, &position, (void *) &tipo_entrada, 1, tipo[MSG_INT], MPI_COMM_WORLD);
    MPI_Unpack(buffer, size, &position, (void *) &tamanho_entrada, 1, tipo[MSG_INT], MPI_COMM_WORLD);
    MPI_Unpack(buffer, size, &position, (void *) &tag, 1, tipo[MSG_INT], MPI_COMM_WORLD);
    
    /*Calcula o tamanho do vetor necessario*/
    MPI_Pack_size(tamanho_entrada, tipo[tipo_entrada], MPI_COMM_WORLD, &tamanho_buffer);
    
    /*Aloca o vetor*/
    vetor = malloc (tamanho_buffer);
    
    /*Desempacota a informacao enviada*/
    MPI_Unpack(buffer, size, &position, (void *) vetor, tamanho_entrada, tipo[tipo_entrada], MPI_COMM_WORLD);

    /*Adiciona o registro na propria classe*/
    this->adiciona_pacote(vetor, tipo_entrada, tamanho_entrada, tag);
  }
  
}
