#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "hmnr.h"
#include "mpi.h"

#define HMNR_DV_TAG 30000
#define HMNR_SIMPLE_TAG 30001
#define HMNR_SENT_TAG 30002
#define HMNR_SYNCH_TAG 30003
#define HMNR_CHECKPOINT_CLOCK_TAG 30004

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo HMNR
  Criada por Raphael Marcos Menderico
  Criacao: 09/12/2003
  Ultima alteracao: 09/12/2003
*/

hmnr::hmnr(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  /*A rede j foi inicializada*/
  int i;
  /*Alocando vetores*/
  
  this->dv_size = this->tamanho_rede();
  this->dv = (int *) calloc (this->dv_size, sizeof(int));
  this->simple = (int *) calloc (this->dv_size, sizeof(int));
  this->sent = (int *) calloc (this->dv_size, sizeof(int));
  this->synch = (int *) calloc (this->dv_size, sizeof(int));
  this->dv_pid = this->id_processo();
  /*Inicializando vetores*/
  
  this->checkpoint_clock = -1;
  
  for (i = 0; i < this->dv_size; i++) {

    this->sent[i] = 0; /*False*/
    this->simple[i] = 0; /*False*/
    this->synch[i] = 0; /*False*/
    this->dv[i] = 0;
  }
  this->simple[this->dv_pid] = 1; /*true*/   
  this->synch[this->dv_pid] = 1; /*true*/   
  this->dv[this->dv_pid]++;
}

hmnr::~hmnr() {
  free(this->dv);
  free(this->simple);
  free(this->sent);
  free(this->synch);
}

int hmnr::processa_recebimento(buffer *b, mensagem_status *status) {

  void *vetor;
  int tipo_p, tamanho_p, tag_p;
  int *dv_msg;
  int *simple_msg;
  int *synch_msg;
  int lc_msg;
  int i;

  int c1, c2;
  
  /*Alocacao de memria*/

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  simple_msg = ((int *) vetor);

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  synch_msg = ((int *) vetor);

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  dv_msg = ((int *) vetor);

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  lc_msg = *((int *) vetor);

  /*Algoritmo de checkpoint forcado*/
  
  if (lc_msg > this->checkpoint_clock) {
    c1 = (dv_msg[this->dv_pid] == this->dv[this->dv_pid]) && 
      (!simple_msg[this->dv_pid]);
    c2 = 0; /*False*/
    
    for (i=0; i<this->dv_size; i++) {
      c2 = c2 || (this->sent[i] && synch_msg[i]);
    }
    if (c1 || c2) {
      this->dv[this->dv_pid]++;
      for (i = 0; i < this->dv_size; i++) {
	this->sent[i] = 0; /*False*/
	if (i != this->dv_pid) {
	  this->simple[i] = 0; /*False*/
	  this->synch[i] = 0; /*False*/
	}
      }
      this->take_checkpoint(1);
    }
    this->checkpoint_clock = lc_msg;
    for (i = 0; i < this->dv_size; i++) {
      if (i != this->dv_pid) {
	this->synch[i] = synch_msg[i]; 
      }
    }
    this->synch[this->dv_pid] = 1 /*True*/;
  } else if (this->checkpoint_clock == lc_msg){
    for (i = 0; i < this->dv_size; i++) {
      this->synch[i] = synch_msg[i] || this->synch[i]; 
    }
  }

  for (i = 0; i < this->dv_size; i++) {
    if (i != this->dv_pid) {
      if (dv_msg[i] > this->dv[i]) {
	this->dv[i] = dv_msg[i];
	this->simple[i] = simple_msg[i]; 
      } else if (dv_msg[i] == this->dv[i]) {
	this->simple[i] = simple_msg[i] && this->simple[i];
      }
    }
  }
  


  


  /*Liberacao de memria*/
  free(synch_msg);
  free(simple_msg);
  free(dv_msg); 

  return(0);

}

void hmnr::informacoes_controle(buffer *b) {
  /*Ao solicitar as informacoes de controle, isso significa que 
    um checkpoint esta sendo tirado pelo sistema*/
  b->adiciona_pacote((void *)(this->dv), MSG_INT, this->dv_size, HMNR_DV_TAG);

  b->adiciona_pacote((void *)&(this->checkpoint_clock), MSG_INT, 
		       1, HMNR_CHECKPOINT_CLOCK_TAG);


}

void hmnr::tira_checkpoint_basico() {
  int i;
  this->dv[this->dv_pid]++;
  this->checkpoint_clock++;

  this->dv[this->monitor()] = -1;


  for (i = 0; i < this->dv_size; i++) {
    this->sent[i] = 0; /*False*/
    if (i != this->dv_pid) {
      this->simple[i] = 0; /*False*/
      this->synch[i] = 0; /*False*/
    }
  }

  mensagem::tira_checkpoint_basico();
}

int hmnr::envia_mensagem(buffer *buf, int dest, int tag) {
  /*Adicionando o relogio a mensagem enviada*/
  
  this->sent[dest] = 1; /*True*/

  buf->adiciona_pacote((void *)&(this->checkpoint_clock), MSG_INT, 
		       1, HMNR_CHECKPOINT_CLOCK_TAG);

  buf->adiciona_pacote((void *)(this->dv), MSG_INT, 
		       this->dv_size, HMNR_DV_TAG);

  buf->adiciona_pacote((void *)(this->synch), MSG_INT, 
		       this->dv_size, HMNR_SYNCH_TAG);

  buf->adiciona_pacote((void *)(this->simple), MSG_INT, 
		       this->dv_size, HMNR_SIMPLE_TAG);
  

  mensagem::envia_mensagem(buf, dest, tag);
  return(0);
}

char* hmnr::algoritmo_utilizado() {
  return ("HMNR\n");
}
void hmnr::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void hmnr::inicializa_processo() {
  mensagem::inicializa_processo();
}

