#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "kt.h"
#include "mpi.h"

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo KT (sincrono bloqueante)
  Criada por Raphael Marcos Menderico
  Criacao: 10/12/2003
  Ultima alteracao: 12/12/2003
*/

#define KT_REQ_TAG 30000
#define KT_RES_TAG 30001
#define KT_LIB_TAG 30002

#define KT_IND_MSG_TAG 30003

kt::kt(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  this->tamanho_vetores = this->tamanho_rede();
  this->ultima_msg_rec = (int *) calloc (this->tamanho_vetores, sizeof(int));
  this->primeira_msg_env = (int *) calloc (this->tamanho_vetores, sizeof(int));
  this->pid = this->id_processo();
  this->num_colaboradores = 0;
  this->blocked = 0; /*False*/
  this->iniciador = 0; /*False*/
}

kt::~kt() {
  free(this->ultima_msg_rec);
  free(this->primeira_msg_env);
}

int kt::verifica_mensagem_controle (int tag) {
  return ((tag == KT_REQ_TAG) || (tag == KT_RES_TAG) || (tag == KT_LIB_TAG));
}

int kt::processa_recebimento(buffer *b, mensagem_status *status) {
  int m_ind_msg;
  int tipo_p, tamanho_p, tag_p;
  void *vetor;
  int i;
  buffer *buf;
  switch (status->tag){
  case KT_REQ_TAG:
    b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
    m_ind_msg = *((int *)vetor);
    if (!this->blocked) {
      if ((this->primeira_msg_env[status->source] > 0) && 
	  (m_ind_msg >= this->primeira_msg_env[status->source])) {
	this->emissor_primeira_req = status->source;
	this->take_checkpoint(2);
	this->num_colaboradores = 0;
	for (i=0; i < this->tamanho_vetores; i ++) {
	  this->primeira_msg_env[i] = 0;
	  if (this->ultima_msg_rec[i] != 0) {
	    this->num_colaboradores++;
	  }
	}
	if (this->num_colaboradores > 0) {
	  this->blocked = 1; /*true*/
	  for (i=0; i < this->tamanho_vetores; i++) {
	    if (this->ultima_msg_rec[i] != 0) {
	      buf = new buffer ();
	      buf->adiciona_pacote((void *)&(this->ultima_msg_rec[i]), MSG_INT, 1, KT_IND_MSG_TAG);
	      mensagem::envia_mensagem(buf, i, KT_REQ_TAG);
	      free(buf);
	    }
	  }    
	} else {
	  buf = new buffer();
	  mensagem::envia_mensagem(buf, this->emissor_primeira_req, KT_RES_TAG);
	  free(buf);
	}
      } else {
	buf = new buffer();
	mensagem::envia_mensagem(buf, status->source, KT_RES_TAG);
	free(b);
      }
    } else {
      buf = new buffer();
      mensagem::envia_mensagem(buf, status->source, KT_RES_TAG);
      free(b);
    } 
    break;
  case KT_RES_TAG:
    this->num_colaboradores--;
    if (this->num_colaboradores == 0) {
      if (this->iniciador) {
	this->blocked = 0; /*False*/
	for (i=0; i < this->tamanho_vetores; i++) {
	  if (this->ultima_msg_rec[i] != 0) {
	    buf = new buffer ();
	    mensagem::envia_mensagem(buf, i, KT_LIB_TAG);
	    free(buf);
	    this->ultima_msg_rec[i] = 0;
	  }
	}    
  
      } else {
	buf = new buffer();
	mensagem::envia_mensagem(buf, this->emissor_primeira_req, KT_RES_TAG);
	free(buf);
      }
    }
    break;
  case KT_LIB_TAG:
    if (this->blocked) {
      this->blocked = 0; /*False*/
      for (i=0; i < this->tamanho_vetores; i++) {
	if (this->ultima_msg_rec[i] != 0) {
	  buf = new buffer ();
	  mensagem::envia_mensagem(buf, i, KT_LIB_TAG);
	  free(buf);
	  this->ultima_msg_rec[i] = 0;
	}
      }    
    }
    break;
  default:
    b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
    m_ind_msg = *((int *)vetor);
    ultima_msg_rec[status->source] = m_ind_msg;
  }
    
  return(0);
}

void kt::informacoes_controle(buffer *b) {
  
}

void kt::tira_checkpoint_basico() {
  buffer *b;
  int i;
  this->iniciador = 1; /*True*/
  mensagem::tira_checkpoint_basico();
  this->num_colaboradores = 0;
  for (i = 0; i < this->tamanho_vetores; i++) {
    if (this->ultima_msg_rec[i] != 0) {
      this->num_colaboradores++;
    }
  }
  if (this->num_colaboradores == 0) {
    for (i=0; i < this->tamanho_vetores; i++) {
      this->primeira_msg_env[i] = 0;
    }
  } else {
    this->blocked = 1; /*True*/
    for (i=0; i < this->tamanho_vetores; i++) {
      if (this->ultima_msg_rec[i] != 0) {
	b = new buffer ();
	b->adiciona_pacote((void *)&(this->ultima_msg_rec[i]), MSG_INT, 1, KT_IND_MSG_TAG);
	mensagem::envia_mensagem(b, i, KT_REQ_TAG);
	free(b);
      }
    }    
  }

  /*Se bloqueado, realizar o bloqueio do recebimento*/
  if (this->blocked) {
    this->recebe_mensagem(1);
  }

}

int kt::envia_mensagem(buffer *buf, int dest, int tag) {
  this->ind_msg++;
  if (this->primeira_msg_env[dest] == 0) {
    this->primeira_msg_env[dest] = this->ind_msg;
  }
  buf->adiciona_pacote((void *)&(this->ind_msg), MSG_INT, 1, KT_IND_MSG_TAG);
  mensagem::envia_mensagem(buf, dest, tag);
  return(1);
}

char* kt::algoritmo_utilizado() {
  return ("Koo-Toueg\n");
}
void kt::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void kt::inicializa_processo() {
  mensagem::inicializa_processo();
}
