#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "lb.h"
#include "mpi.h"

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo Leu-Bhargava (sincrono bloqueante)
  Criada por Raphael Marcos Menderico
  Criacao: 20/01/2004
  Ultima alteracao: 21/01/2004
*/

#define LB_REQ_TAG 30000
#define LB_RES_TAG 30001
#define LB_LIB_TAG 30002
#define LB_CONF_TAG 30003

#define LB_IND_MSG_TAG 30004
#define LB_FILHO_TAG 30005

#define LB_NIL -1

lb::lb(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  int i;
  this->tamanho_vetores = this->tamanho_rede();
  this->vi_ckpt = (int *) calloc (this->tamanho_vetores, sizeof(int));
  this->filhos = (int *) calloc (this->tamanho_vetores, sizeof(int));
  this->pid = this->id_processo();

  for (i=0; i < this->tamanho_vetores; i++) {
    this->vi_ckpt[i] = 0;
    this->filhos[i] = 0; /*Falso*/
  }
  this->pai = LB_NIL;
  this->num_filhos = 0;
  this->cont_req = 0;
  this->vi_ckpt[this->pid]++;
  this->blocked = 0; /*False*/
}

lb::~lb() {
  free(this->vi_ckpt);
  free(this->filhos);
}

int lb::verifica_mensagem_controle (int tag) {
  return ((tag == LB_REQ_TAG) || (tag == LB_RES_TAG) 
	  || (tag == LB_LIB_TAG) || (tag == LB_CONF_TAG));
}

int lb::processa_recebimento(buffer *b, mensagem_status *status) {
  int m_ind_msg;
  int filho_msg;
  int tipo_p, tamanho_p, tag_p;
  void *vetor;
  int i;
  /*Variaveis para serem enviadas como confirmao*/
  int falso = 0;    
  int verdadeiro = 1;
  buffer *buf;
  switch (status->tag){
  case LB_REQ_TAG:
    b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
    m_ind_msg = *((int *)vetor);
    if (this->blocked) {
      buf = new buffer ();
      buf->adiciona_pacote((void *)&(falso), MSG_INT, 1, LB_FILHO_TAG);
      mensagem::envia_mensagem(buf, status->source, LB_CONF_TAG);
      free(buf);
    } else {
      if (this->vi_ckpt[this->pid] <= m_ind_msg) {
	buf = new buffer ();
	buf->adiciona_pacote((void *)&(verdadeiro), MSG_INT, 1, LB_FILHO_TAG);
	mensagem::envia_mensagem(buf, status->source, LB_CONF_TAG);
	free(buf);
	this->pai = status->source;
	this->vi_ckpt[this->pid]++;
	this->take_checkpoint(2);
	this->cont_req = 0;
	for (i=0; i < this->tamanho_vetores; i++) {
	  if ((this->vi_ckpt[i] != 0) && (i != this->pid) && 
	      (i != this->pai) && (i != this->monitor())) {
	    this->cont_req++;
	    buf = new buffer ();
	    buf->adiciona_pacote((void *)&(this->vi_ckpt[i]), 
				 MSG_INT, 1, LB_IND_MSG_TAG);
	    mensagem::envia_mensagem(buf, i, LB_REQ_TAG);
	    free(buf);
	  }
	}
	if (this->cont_req == 0) {
	  buf = new buffer ();
	  mensagem::envia_mensagem(buf, this->pai, LB_RES_TAG);
	  free(buf);
	} else {
	  this->blocked = 1; /*True*/
	}
      } else {
	buf = new buffer ();
	buf->adiciona_pacote((void *)&(falso), MSG_INT, 1, LB_FILHO_TAG);
	mensagem::envia_mensagem(buf, status->source, LB_CONF_TAG);
	free(buf);
      }
    }
    break;
  case LB_CONF_TAG:
    b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
    filho_msg = *((int *)vetor);
    if (filho_msg) {
      this->num_filhos++;
      this->filhos[status->source] = 1; /*True*/
    }
    this->cont_req--;
    if ((cont_req == 0) && (this->num_filhos == 0)) {
      if (this->pai == LB_NIL) {
	this->blocked = 0; /*False*/
      } else {
	buf = new buffer ();
	mensagem::envia_mensagem(buf, this->pai, LB_RES_TAG);
	free(buf);
      }
    }
    break;
  case LB_RES_TAG:
    this->num_filhos--;
    if (this->num_filhos == 0) {
      if (this->pai == LB_NIL) {
	/*Libera todos os filhos*/
	this->blocked = 0; /*False*/
	for (i = 0; i < this->tamanho_vetores; i++) {
	  if ((this->filhos[i]) && (i != this->pid) 
	      && (i != this->monitor())) {
	    buf = new buffer();
	    mensagem::envia_mensagem(buf, i, LB_LIB_TAG);
	    free(buf);	  
	    this->vi_ckpt[i] = 0;
	    this->filhos[i] = 0;
	  }
	}
      } else { 
	/*Transmite resposta para o pai*/
	buf = new buffer();
	mensagem::envia_mensagem(buf, this->pai, LB_RES_TAG);
	free(buf);	  
	
      }
    }
    break;
  case LB_LIB_TAG:
    if (this->blocked) {
      this->blocked = 0; /*False*/
      for (i=0; i < this->tamanho_vetores; i++) {
	if (filhos[i] && (i != this->pid) && (i != this->monitor())) {
	  buf = new buffer();
	  mensagem::envia_mensagem(buf, i, LB_LIB_TAG);
	  free(buf);	  
	  this->vi_ckpt[i] = 0;
	  this->filhos[i] = 0;
	}
      }
    }
    break;
  default:
    b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
    m_ind_msg = *((int *)vetor);
    this->vi_ckpt[status->source] = m_ind_msg;
  }
    
  return(0);
}

void lb::informacoes_controle(buffer *b) {
  
}

void lb::tira_checkpoint_basico() {
  buffer *buf;
  int i;

  this->vi_ckpt[this->pid]++;
  this->take_checkpoint(0);
  for (i=0; i < tamanho_vetores; i++) {
    for (i=0; i < this->tamanho_vetores; i++) {
      if ((this->vi_ckpt[i] != 0) && (i != this->pid) && 
	  (i != this->pai) && (i != this->monitor())) {
	this->cont_req++;
	buf = new buffer ();
	buf->adiciona_pacote((void *)&(this->vi_ckpt[i]), 
			     MSG_INT, 1, LB_IND_MSG_TAG);
	mensagem::envia_mensagem(buf, i, LB_REQ_TAG);
	free(buf);
	this->pai = LB_NIL;
	this->blocked = 1; /*True*/
      }
    }
  }

  /*Se bloqueado, realizar o bloqueio do recebimento*/
  if (this->blocked) {
    this->recebe_mensagem(1);
  }

}

int lb::envia_mensagem(buffer *buf, int dest, int tag) {

  buf->adiciona_pacote((void *)&(this->vi_ckpt[this->pid]), MSG_INT, 1, LB_IND_MSG_TAG);
  mensagem::envia_mensagem(buf, dest, tag);
  return(1);

}

char* lb::algoritmo_utilizado() {
  return ("Leu-Bhargava\n");
}
void lb::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void lb::inicializa_processo() {
  mensagem::inicializa_processo();
}
