#include "mensagem.h"
#include "bcs.h"
#include "iostream.h"
#include "mpi.h"

#define TOKEN_TAG 1
#define STEP_TAG 2
#define DIE_TAG 3
#define SEND_TAG 4
#define LIMITE_PASSADAS 500

int token = 0;
int contador_passadas = 0;

void carrega_estado (buffer *estado) {
  /*Carregando um estado teletubbie*/
  int *dummy ;
  dummy = (int *) malloc (sizeof(int));
  (*dummy) = token;
  estado->adiciona_pacote((void *)dummy, MSG_INT, 1, TOKEN_TAG);
  
  dummy = (int *) malloc (sizeof(int));
  (*dummy) = contador_passadas;
  estado->adiciona_pacote((void *)dummy, MSG_INT, 1, STEP_TAG);
}

int main (int argc, char* argv[]) {
  
  mensagem* M;
  buffer *B;
  mensagem_status s;
  int proximo_processo;
  void *vetor;
  int tag_p;
  int tipo_p;
  int tamanho_p;
  int conta_processos;
  int i;

  M = new bcs(&argc, &argv, carrega_estado);
  B = NULL;
  M->inicializa_processo();

  /*Define o proximo processo no anel de tokens
    isso eh sempre constante*/

  proximo_processo = ((M->id_processo() + 1) % M->tamanho_rede());
  if (proximo_processo == M->monitor()) { 
      proximo_processo = ((proximo_processo + 1) % M->tamanho_rede());
  }

  if (M->id_processo() == 1) {
    /*Primeiro processo - aguarda o desbloqueio e gera um token
      Como nao foi implementado nenhum algoritmo de eleicao de lider
      ou coisa parecida, vai ser arbitrario mesmo 
      Poderia ser um codigo fonte diferente, mas .. nao tem 
      muita necessidade*/
    M->aguarda_desbloqueio();

    token = 1;
    contador_passadas = 1;

    B = new buffer();

    B->adiciona_pacote((void *)&token, MSG_INT, 1, TOKEN_TAG);
    B->adiciona_pacote((void *)&contador_passadas, MSG_INT, 1, TOKEN_TAG);    
    M->envia_mensagem(B,proximo_processo,0);
  }

  do {
    M->recebe_mensagem(&B, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    if (s.tag != DIE_TAG) {

      /*Recolhendo informacoes da mensagem recebida*/
      B->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
      contador_passadas = *((int *)vetor);
      free(vetor);
      B->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
      token = *((int *)vetor);      
      free(vetor);
      delete(B);

      /*Atualizando informacoes sobre os pacotes*/
      contador_passadas++;
      
      if (contador_passadas <= LIMITE_PASSADAS) { 
	M->tira_checkpoint_basico();      
	/*Criando novo buffer para envio da proxima mensagem*/
	B = new buffer();
	B->adiciona_pacote((void *)&token, MSG_INT, 1, TOKEN_TAG);
	B->adiciona_pacote((void *)&contador_passadas, MSG_INT, 1, TOKEN_TAG);    
	/*Repassando o token*/
	M->envia_mensagem(B,proximo_processo,SEND_TAG);      
	/*Tirando um checkpoint basico */
	token = 0;
	M->tira_checkpoint_basico();
      } else {
	/*Caso contrario, enviando uma mensagem de morte para todos 
	  os processos e depois saindo do laco*/
	conta_processos = M->tamanho_rede();
	B = new buffer();
	for (i=0; i < conta_processos; i++) {
	  if ((i != M->id_processo()) && (i != M->monitor()))
	  M->envia_mensagem(B, i, DIE_TAG);
	}
      }
    }
  } while (s.tag != DIE_TAG && contador_passadas <= LIMITE_PASSADAS);
  M->finaliza_processo();
  return (0);
}
