#include <stdio.h> 
#include <iostream.h>
#include <string.h>
#include "mensagem.h"
#include "mpi.h"
#include "constantes.h"

#define STATE_BLOCKED 1
#define STATE_WAITING_RELEASE 2
#define STATE_RELEASED 3

#define TYPE_WAIT_RELEASE_MESSAGE 0
#define TYPE_DONT_WAIT_RELEASE_MESSAGE 1

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Superclasse abstrata de envio e recebimento de mensagens
  Criada por Raphael Marcos Menderico
  Criacao: 01/07/2003
  Ultima alteracao: 11/12/2003
*/

mensagem:: mensagem(int *argc, char** argv[], void (*grava_estado) (buffer *estado)){
  this->fila = new fifo();
  this->coleta_estado = grava_estado;
  this->logical_clock = -1;
  this->checkpoint_count = -1;
  this->rank = -1;
  MPI_Init (argc, argv);
}

mensagem::~mensagem() {
  MPI_Finalize();
}

int mensagem::verifica_mensagem_controle (int tag) {
  return (0); /*False*/
}

int mensagem::send(buffer *b, int dest, int tag) {
  /*Funcao de envio de mensagens
    Atualmente utilizando o LAM-MPI*/
  void *buffer = NULL;
  int size = 0;

  b->empacota(&buffer, &size);

  MPI_Send (buffer, size, MPI_PACKED, dest, tag, MPI_COMM_WORLD);

  free(buffer);

  return (0);
}

int mensagem::timestamp() {
  /*Funcao para depuracao do codigo*/
  return (this->logical_clock);
}

int mensagem::receive(buffer **buf, int dest, int tag, mensagem_status *status) {

  void *buf_temp = NULL;
  int tamanho = 0;

  int dest_mpi = MPI_ANY_SOURCE;
  int tag_mpi = MPI_ANY_TAG;

  MPI_Status status_mpi;
  
  /*Atualizando o valor do destino e do tag*/
  if (dest != MSG_ANY_SOURCE) { 
    dest_mpi = dest;
  }
  
  if (tag != MSG_ANY_TAG) { 
    tag_mpi = tag;
  }

  /*Verificando o tamanho da mensagem*/
  MPI_Probe(dest_mpi, tag_mpi, MPI_COMM_WORLD, &status_mpi);

  tamanho = status_mpi.st_length;

  /*Alocando o buffer necessário*/
  buf_temp = malloc (tamanho);

  /*Recebendo a mensagem*/
  MPI_Recv (buf_temp, tamanho, MPI_PACKED, dest_mpi, tag_mpi, MPI_COMM_WORLD, &status_mpi);

  (*buf) = new buffer();
  (*buf)->desempacota (buf_temp, tamanho);

  free(buf_temp);

  /*Atualizando o status (retorno)*/
  status->source = status_mpi.MPI_SOURCE;
  status->tag = status_mpi.MPI_TAG;
  status->size = tamanho; /*Tamanho do pacote original - somente para controle*/
  
  return (0);
}

int mensagem::monitora(buffer **buf, int dest, int tag, mensagem_status *status){
  return(this->receive(buf, dest, tag, status));
}

int mensagem::tamanho_rede() {
  int size;
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  return(size);
}

int mensagem::id_processo() {
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  return(rank);
}

int mensagem::monitor () {
  /*Funcao incluida para fins de depuracao*/
  /*Retorna o rank do monitor*/
  return (this->rank);
}

int mensagem::envia_mensagem(buffer *buf, int dest, int tag) {
  
  /*Envia a mensagem somente se estiver desbloqueado ou se 
   for uma mensagem de controle, caso contrrio, descarta o envio*/
  if ((!this->blocked) || this->verifica_mensagem_controle(tag)) {
  /*Incrementa o relogio logico*/
    this->logical_clock++;
    
    /*Adiciona o relogio ao buffer*/
    
    buf->adiciona_pacote((void *) & (this->logical_clock), MSG_INT, 1, MSG_LOGICAL_CLK_TAG);
    
    this->send(buf, dest, tag);
  }
  return(0);
}

int mensagem::recebe_mensagem(buffer **buf, int source, int tag, mensagem_status* status) {

  return (mensagem::recebe_mensagem(buf, source, tag, status, 
				    TYPE_WAIT_RELEASE_MESSAGE));
}

int mensagem::recebe_mensagem(int tipo) {
  buffer *buf;
  mensagem_status status;

  return (mensagem::recebe_mensagem(&buf, MSG_ANY_SOURCE, MSG_ANY_TAG, 
				    &status, TYPE_DONT_WAIT_RELEASE_MESSAGE));

}
int mensagem::recebe_mensagem(buffer **buf, int source, int tag, mensagem_status* status, int tipo) {
  int tipo_p, tamanho_p, tag_p, *relogio_p;
  
  Mensagem_fila *m;

  /*Tipo TYPE_WAIT_RELEASE_MESSAGE -> aguarda uma mensagem valida aps um
   desbloqueio.*/

  buffer *b = NULL; 

  int estado = -1;
  
  /*Determinando o estado inicial*/
  if (!this->blocked) {
    estado = STATE_WAITING_RELEASE;
  } else {
    estado = STATE_BLOCKED;
  }

  do {
    do {

      /*Alterar depois - qdo estiver em estado de bloqueio, receber 
       qq tipo de mensagem, e nao somente aquelas especificadas pelo usuario,
       talvez retirar essa funcionalidade seja uma ideia .. */
      
      /*Verifica se no existem mensagens na fila para serem processadas,
	se existirem, processa elas antes */
      if (this->fila->tamanho_atual() == 0 || this->blocked) {
	this->receive(buf, source, tag, status);
      } else {
	m = (Mensagem_fila *) (this->fila->remove());
	(*buf) = m->buf;
	status->source = m->source;
	status->tag = m->tag;
	status->size = m->size;
	free(m);
      }
      if (status->tag == MSG_RELEASE_TAG) {
	free(*buf);
      }
    } while (status->tag == MSG_RELEASE_TAG); /*Despreza uma mensagem
						enquanto nao receber
						um tag que nao seja de release*/
    
    /*Se no estiver bloqueado ou se for uma mensagem de controle, processa
      imediatamente a mensagem*/
    if ((!this-> blocked) || (this->verifica_mensagem_controle(status->tag))) {
      /*Removendo o relogio logico*/
      (*buf)->remove_pacote((void **)&relogio_p, &tipo_p, &tamanho_p, &tag_p);
      if (tag_p == MSG_LOGICAL_CLK_TAG) {
	/*Atualiza o relogio logico*/
	if ((*relogio_p) >= this->logical_clock) { 
	  this->logical_clock = (*relogio_p); 
	}
	
	this->logical_clock++;
	
	/*Enviando mensagem ao monitor*/
	b = new buffer();
	b->adiciona_pacote((void *)&(this->logical_clock), MSG_INT, 1, 
			   MSG_DEST_CLOCK_TAG);
	b->adiciona_pacote((void *)relogio_p, MSG_INT, 1, 
			   MSG_SOURCE_CLOCK_TAG);		       

	/*Nova informacao enviada - tag da mensagem*/
	b->adiciona_pacote((void *)&(status->tag), MSG_INT, 1,
			   MSG_TAG_TAG);

	b->adiciona_pacote((void *)&(status->source), MSG_INT, 1,
			   MSG_SOURCE_TAG);


	this->send(b, this->rank, MSG_MONITOR_MSG_TAG);
	delete(b);	       

	free(relogio_p);

	/*Processa o recebimento da mensagem pelo protocolo*/
	this->processa_recebimento(*buf, status);
	

      }
    } else {
      /*Caso contrrio (est bloqueado e a mensagem recebida no  de controle)
	Armazena a mensagem na fila*/
      m = (Mensagem_fila *) calloc (1, sizeof(Mensagem_fila));
      m->buf = (*buf);
      m->source = status->source;
      m->tag = status->tag;
      m->size = status->size;
      fila->adiciona((void *) m);
    }

    if (this->blocked) {
      estado = STATE_BLOCKED;
    } else {
      if (estado == STATE_BLOCKED) {
	if (tipo == TYPE_WAIT_RELEASE_MESSAGE) {
	  estado = STATE_WAITING_RELEASE; 
	} else {
	  estado = STATE_RELEASED; 
	}
      } else if (estado == STATE_WAITING_RELEASE) {
	if (!this->verifica_mensagem_controle(status->tag)) {
	  estado = STATE_RELEASED; 
	}
      }
    }
  } while (estado != STATE_RELEASED);
  return (0);
}
char* mensagem::algoritmo_utilizado() {
  return ("Superclasse");
}

void mensagem::inicializa_monitor() {
  this->inicializa_monitor(NULL);
}

void mensagem::inicializa_monitor(buffer *** retorno) {
  /*Durante a inicializacao do monitor, ele deve pegar e enviar uma mensagem
    para todos os demais processos avisando qual eh o seu rank - pelo menos 
    no lam-mpi eh assim - caso mude, a inicializacao muda :-) - Retorna o 
    checkpoint inicial, se um ponteiro para buffer for passado*/
  
  buffer *b;
  buffer *temp;
  int numero_processos, id_monitor;
  mensagem_status s;
  int i;

  numero_processos = this->tamanho_rede();
  id_monitor = this->id_processo();

  b = new buffer();

  b->adiciona_pacote((void *) &id_monitor, MSG_INT, 1, MSG_MONITOR_RANK);

  for (i=0; i<numero_processos; i++) {
    if (i != id_monitor) {
      this->send(b, i, MSG_INIT_TAG); 
    }
  }
  delete(b);

  /*Aguardando confirmacao dos processos*/
  i = (numero_processos - 1);
  while(i > 0) {
    this->receive(&temp, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    /*A confirmacao vem com o checkpoint inicial*/
    if (s.tag == MSG_MONITOR_CHK_TAG) {
      i--;
      if (retorno != NULL) {
	/*Adiciona o numero do processo no cabecalho*/
	/*temp->adiciona_pacote((void *)&(s.source), MSG_INT, 1, MSG_SOURCE_TAG);*/
	/*Coloca a mensagem no buffer*/
	(*retorno)[i] = temp;
      } else {
	delete(temp);
      }
    }
  } 
  /*Enviando mensagem de liberacao
    
    Essa mensagem deve ser tratada individualmente por cada processo, 
    conforme a aplicacao. Basicamente, se um processo deve RECEBER uma mensagem, basta ignorar as mensagens de RELEASE, se ele for enviar, ele deve aguardar uma mensagem de RELEASE e so depois iniciar o processamento.*/

  b = new buffer();
  
  for (i=0; i<numero_processos; i++) {
    if (i != id_monitor) {
      this->send(b, i, MSG_RELEASE_TAG); 
    }
  }
  delete(b);


}
void mensagem::inicializa_processo() {
  /* 
     A inicializacao do processo aguarda a primeira mensagem enviada pelo monitor para comecar o seu processamento - nao eh possivel fazer nada sem saber quem eh o processo pai
  */
  buffer *b = NULL;
  mensagem_status s;
  int *msg_p; 
  int tamanho_p, tipo_p, tag_p;

  do {
    /*Todas as mensagens q nao forem de inicializacao serao ignoradas -
      em tese, os processos deveriam estar bloqueados*/
    this->receive(&b, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    if (s.tag == MSG_INIT_TAG) {
      b->remove_pacote((void **)&msg_p, &tipo_p, &tamanho_p, &tag_p);
      if (tag_p == MSG_MONITOR_RANK) {
	this->rank = (*msg_p);
      }
    }
    delete(b);
  } while (s.tag != MSG_INIT_TAG);
  
  this->take_checkpoint(0);

  /*Envia a mensagem de confirmacao*/
  

}

void mensagem::aguarda_desbloqueio() {
  /*Para processos que devem enviar mensagens logo depois
    de inicializados, eh necessario aguardar o envio de uma mensagem
    do monitor, desbloqueando o processo. os demais
    podem simplesmente ignorar as mensagens com o tag MSG_RELEASE_TAG, ja 
    que devem receber mensagens logo apos a inicializacao (e portanto, o 
    bloqueio eh desnecessario*/
  /*Os processos que devem esperar essa mensagem ignorarao TODAS as mensagens
    que chegarem antes do desbloqueio, portanto, fica a cargo da aplicacao
    nao enviar mensagens a esses processos, se nao quiser que elas sejam 
    ignoradas*/

  buffer *b = NULL;
  mensagem_status s;

  do {
    /*Todas as mensagens q nao forem de inicializacao serao ignoradas -
      em tese, os processos deveriam estar bloqueados*/
    this->receive(&b, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    delete(b);
   } while (s.tag != MSG_RELEASE_TAG);


}

void mensagem::finaliza_processo() {
  /*
    Todo processo deve enviar para o monitor, antes de morrer, uma mensagem
    comunicando que esta parando de executar. O monitor registrara o timestamp
    da mensagem e nao aceitara mensagens posteriores de monitoramento do 
    processo em questao.
   */

  buffer *b;
  int relogio;

  b = new buffer();

  /*Tira o checkpoint basico final*/
  this->take_checkpoint(0);

  relogio = this->logical_clock;
  
  b->adiciona_pacote((void *)&relogio, MSG_INT, 1, MSG_LOGICAL_CLK_TAG);

  this->send(b, this->rank, MSG_MONITOR_DIE_TAG);

  delete(b);
  
}

/*Funcao que envia o checkpoint para o monitor. 
  O flag indica se o checkpoint eh basico ou nao*/
int mensagem::take_checkpoint(int forced) {
  /*Cria o novo buffer*/
  buffer *b;
  int eu = this->id_processo();
  b = new buffer();
  
  /*Coleta o estado atual*/
  (*(this->coleta_estado)) (b);
  
  /*Adiciona uma informacao de controle do proprio algoritmo*/
  this->informacoes_controle(b);

  /*Adiciona ao estado coletado o contador de checkpoints, o relogio
    atual (ambos incrementados de 1) e um booleano indicando se eh 
    um checkpoint forcado ou nao*/
  if (forced == 0 || forced == 2) {
    (this->logical_clock)++;
  }
  (this->checkpoint_count)++;
  
  b->adiciona_pacote((void *)&(this->logical_clock), 
		     MSG_INT, 1, MSG_LOGICAL_CLK_TAG);
  b->adiciona_pacote((void *)&(this->checkpoint_count), 
		     MSG_INT, 1, MSG_CHECKPOINT_COUNT_TAG);
  b->adiciona_pacote((void *)&(forced), 
		     MSG_INT, 1, MSG_FORCED_TAG);
  b->adiciona_pacote((void *)&(eu), 
		     MSG_INT, 1, MSG_SOURCE_TAG);
  this->send(b, this->rank, MSG_MONITOR_CHK_TAG);

  delete(b);

  return(0);

}

void mensagem::tira_checkpoint_basico() {
  this->take_checkpoint(0);
}

void mensagem::informacoes_controle(buffer *b) {
  /*
    Funcao vazia que adiciona ao estado enviado ao monitor
    informacoes que o protocolo deseje adicionar (tipo 
    um relogio vetorial)
   */
}

int mensagem::processa_recebimento(buffer *b, mensagem_status *status) {
  /*Funcao vazia para ser sobrescrita pelos filhos com as 
    atitudes que dizem se um objeto deve ou nao tirar um 
    checkpoint forcado*/
  return(0);
}
