#include "mensagem.h"
#include "bcs.h"
#include "iostream.h"
#include "buffer.h"
#include "fstream.h"

/*Programa de monitoramento dos estados
  Criado por Raphael Marcos Menderico
  Em 29 de julho de 2003
  Ultima alteracao - 29/07/2003*/

/*Funcao q copia o estado atual para um buffer que possa ser enviado
  no processo monitor, serve somente para passar de parametro durante
  a inicializacao*/
void carrega_estado (buffer *estado) {
  
}

int main (int argc, char* argv[]) {
  
  /*Variaveis de controle do monitoramento*/
  mensagem *m;
  buffer *b = NULL;
  buffer **inicio = NULL;
  mensagem_status s;
  int contador_vivos;
  int id_monitor;
  int *hash_processos;

  /*Variaveis de recebimento de mensagens*/
  void *vetor;
  int tipo_p, tamanho_p, tag_p; 
  int relogio;
  int forcado;
  int contador_checkpoint;
  int relogio_origem;
  int origem;
  /*Constroi o arquivo de saida*/
  ofstream saida(argv[1]);
  /*Variaveis gerais*/
  int i, j;

  m = new mensagem(&argc, &argv, carrega_estado);
 
  id_monitor = m->id_processo();
  
  /*Inicializa o contador de vivos com o tamanho atual da rede
    Os processos devem enviar uma mensagem de morte para o monitor
    Para que ele saiba que os processos morreram
    Lembrar que o monitor nao deve entrar na contagem*/
  contador_vivos = (m->tamanho_rede() - 1);

  /*Criando e numerando os processos
    A numeracao dada aos processos nao corresponde necessariamente ao 
    rank de cada um ... */
  hash_processos = (int *) calloc (contador_vivos, sizeof(int));

  j=0;
  for (i=0; i<m->tamanho_rede(); i++) {
    if (i != id_monitor) {
      hash_processos[i] = j;
      (saida) << "PROC " << j << "\n";
      j++;
    }
  }
       
  /*Inicializa o monitor, enviando as informacoes de controle aos processos
    No caso, essa funcao escreve os primeiros checkpoints*/

  inicio = (buffer **) calloc(contador_vivos, sizeof(buffer*));

  m->inicializa_monitor(&inicio);

  /*Se o monitor retornar algo util, entaum remove o checkpoint e o imprime
    Trata-se do primeiro checkpoint*/
  if (inicio != NULL) {

    for(i=0; i<contador_vivos;i++) {
      
      b = inicio[i];
      
      if (b != NULL) {
	
	origem = 0;
	
	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	origem = *((int *) (vetor));
	free(vetor);

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	forcado = *((int *) (vetor));
	free(vetor);       

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	contador_checkpoint = *((int *) (vetor));
	free(vetor);	
	
	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	relogio = *((int *) (vetor));
	free(vetor);	
	
	(saida) << "CHK " << hash_processos[origem] << " " << 
	  contador_checkpoint << " " << relogio << " ";
	if (forcado) {
	  (saida) << "TRUE ";
	} else {
	  (saida) << "FALSE ";
	}
	(saida) << b->tamanho_atual() << " ";
	(saida) << "\n";
	
	while (b->tamanho_atual() > 0) {
	  b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	  (saida) << "ESTADO "  << tag_p << " "
	       << tamanho_p << " ";
	  for (j = 0; j < tamanho_p; j++) {
	    switch(tipo_p) {
	    case MSG_INT:
	      (saida) << ((int *)vetor)[j];
	      break;
	    }
	  }
	  (saida) << "\n"; 
	  free(vetor);
	}
	
	delete(b);
      }
    }
  }

  free(inicio);

  while (contador_vivos > 0) {
    /*Enquanto ainda houver processos rodando, monitora as mensagens enviadas*/
    m->monitora (&b, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    if (s.tag > MSG_MONITOR_TAG) { 
      switch(s.tag) {
      case MSG_MONITOR_MSG_TAG:
	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	origem = *((int *) (vetor));
	free(vetor);	

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	relogio_origem = *((int *) (vetor));
	free(vetor);	

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	relogio = *((int *) (vetor));
	free(vetor);	

	(saida) << "MSG " << hash_processos [origem] << " " << 
	  hash_processos[s.source] << " " << 
	  relogio_origem << " " << relogio << "\n";

	break;
      case MSG_MONITOR_CHK_TAG:

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	origem = *((int *) (vetor));
	free(vetor);

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	forcado = *((int *) (vetor));
	free(vetor);	

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	contador_checkpoint = *((int *) (vetor));
	free(vetor);	

	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	relogio = *((int *) (vetor));
	free(vetor);	

	(saida) << "CHK " << hash_processos[origem] << " " << 
	  contador_checkpoint << " " << relogio << " ";
	if (forcado) {
	  (saida) << "TRUE ";
	} else {
	  (saida) << "FALSE ";
	}
	(saida) << b->tamanho_atual() << " ";
	(saida) << "\n";

	while (b->tamanho_atual() > 0) {
	  b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	  (saida) << "ESTADO "  << tag_p << " "
	       << tamanho_p << " ";
	  for (i = 0; i < tamanho_p; i++) {
	    switch(tipo_p) {
	    case MSG_INT:
	      (saida) << ((int *)vetor)[i];
	      break;
	    }
	  }
	  (saida) << "\n"; 
	  free(vetor);
	}
	
	break;
      case MSG_MONITOR_DIE_TAG:
	/*A mensagem contem um relogio logico*/
	b->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	relogio = *((int *) (vetor));
	(saida) << "MORTE " << hash_processos[s.source] << " " << relogio  << "\n";
	contador_vivos--;
	free(vetor);
	break;
      } 
    } else {
      /*Se nao for um tag do monitor, emitir uma mensagem de erro*/
      (saida) << "#ERRO! Mensagem invalida enviada ao monitor - TAG " << 
	s.tag << "\n";
    }
    delete(b);
  }
   
  free(hash_processos);	 

  delete(m);
}
