#include <stdio.h> 
#include <iostream.h>
#include "mensagem.h"
#include "rdtpartner.h"
#include "mpi.h"

#define RDTPARTNER_DV_TAG 30000
#define RDTPARTNER_SIMPLE_TAG 30001
#define RDT_PARTNER_NO_PARTNER -1
#define RDT_PARTNER_MORE_THAN_ONE_PARTNER -2

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
  Classe implementando o algoritmo RDT-Partner
  Criada por Raphael Marcos Menderico
  Criacao: 27/11/2003
  Ultima alteracao: 27/11/2003
*/

rdtpartner::rdtpartner(int *argc, char** argv[], void (*grava_estado) (buffer *estado)):mensagem::mensagem(argc, argv, grava_estado){
  /*A rede j foi inicializada*/
  int i;
  this->dv_size = this->tamanho_rede();
  this->dv = (int *) calloc (dv_size, sizeof(int));
  this->simple = (int *) calloc (dv_size, sizeof(int));
  this->dv_pid = this->id_processo();
  for (i = 0; i < this->dv_size; i++) {
    this->dv[i] = 0;
    this->simple[i] = 0; /*False*/
  }
  this->dv[this->dv_pid]++;
  this->partner = RDT_PARTNER_NO_PARTNER;
  this->simple[this->dv_pid] = 1; /* True */
}

rdtpartner::~rdtpartner() {
  free(this->dv);
  free(this->simple);
}

int rdtpartner::processa_recebimento(buffer *b, mensagem_status *status) {

  void *vetor;
  int tipo_p, tamanho_p, tag_p;
  int *dv_msg;
  int simple_msg;
  int i;

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  simple_msg = ((int) *((int*) vetor));

  b->remove_pacote (&vetor, &tipo_p, &tamanho_p, &tag_p);
  dv_msg = ((int *)vetor);
  
  if ((this->dv[status->source] < dv_msg[status->source])) {
    if (
	(this->partner != RDT_PARTNER_NO_PARTNER) && 
	(
	 (partner != status->source) || 
	 (
	  (partner == status->source) && 
	  (dv_msg[this->dv_pid] == this->dv[this->dv_pid]) &&
	  (!simple_msg)
	  )
	 )
	) {
      for (i=0; i < this->dv_size; i++) {
	this->simple[i] = 0; /*False*/
      }
      this->dv[this->dv_pid]++;
      this->simple[this->dv_pid] = 1; /*True*/
      this->partner = RDT_PARTNER_NO_PARTNER;
      this->take_checkpoint(1);
    }
    this->simple[status->source] = 1; /*True*/
  }
  for (i=0; i < this->dv_size; i++) {
    if (dv_msg[i] > this->dv[i]) {
      this->dv[i] = dv_msg[i];
    }
  }

  free(dv_msg); 

  return(0);
}

void rdtpartner::informacoes_controle(buffer *b) {
  /*Ao solicitar as informacoes de controle, isso significa que 
    um checkpoint esta sendo tirado pelo sistema*/
  b->adiciona_pacote((void *)(this->dv), MSG_INT, this->dv_size, RDTPARTNER_DV_TAG);
}

void rdtpartner::tira_checkpoint_basico() {
  int i;
  this->dv[this->dv_pid]++;
  this->partner = RDT_PARTNER_NO_PARTNER;
  for (i = 0; i < this->dv_size; i++) {
    if (i == this->monitor()) {
      this->simple[i] = -1;
    } else {
      this->simple[i] = 0; /*False*/
    }
  }
  this->simple[this->dv_pid] = 1; /* True */
  mensagem::tira_checkpoint_basico();
}

int rdtpartner::envia_mensagem(buffer *buf, int dest, int tag) {
  /*Adicionando o relogio a mensagem enviada*/
  
  if (this->partner == RDT_PARTNER_NO_PARTNER) {
    this->partner = dest;
  } else {
    if (this->partner != dest) {
      this->partner = RDT_PARTNER_MORE_THAN_ONE_PARTNER;
    }
  }

  buf->adiciona_pacote((void *)(this->dv), MSG_INT, this->dv_size, RDTPARTNER_DV_TAG);
  buf->adiciona_pacote((void *)&(this->simple[dest]), MSG_INT, 1, RDTPARTNER_SIMPLE_TAG);
  mensagem::envia_mensagem(buf, dest, tag);
  return(0);
}

char* rdtpartner::algoritmo_utilizado() {
  return ("RDT-Partner\n");
}
void rdtpartner::inicializa_monitor() {
  mensagem::inicializa_monitor();
}
void rdtpartner::inicializa_processo() {
  mensagem::inicializa_processo();
}

