#include "mensagem.h"
#include "stdlib.h"
#include "kt.h"
#include "iostream.h"
#include "mpi.h"

#define TOKEN_TAG 1
#define STEP_TAG 2
#define DIE_TAG 3
#define SEND_TAG 4
#define RANDOM_TAG 5
#define LIMITE_PASSADAS 100

/*
 *
 *   This file is part of Lupa
 *
 *   Lupa is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Lupa tributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Foobar; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


int token = 0;
int contador_passadas = 0;

void carrega_estado (buffer *estado) {
  /*Carregando um estado teletubbie*/
  int *dummy ;
  dummy = (int *) malloc (sizeof(int));
  (*dummy) = token;
  estado->adiciona_pacote((void *)dummy, MSG_INT, 1, TOKEN_TAG);
  
  dummy = (int *) malloc (sizeof(int));
  (*dummy) = contador_passadas;
  estado->adiciona_pacote((void *)dummy, MSG_INT, 1, STEP_TAG);
}

int main (int argc, char* argv[]) {
  
  mensagem* M;
  buffer *B;
  mensagem_status s;
  int proximo_processo;
  void *vetor;
  int tag_p;
  int tipo_p;
  int tamanho_p;
  int conta_processos;
  int i;
  /*Vetor para indicar quantos processos estao mortos*/
  int *mortos;
  int morte_iniciada = 0;
  int contador_vivos;
  int mensagens_aleatorias;
  int destino_aleatorio;

  M = new kt (&argc, &argv, carrega_estado);
  B = NULL;
  M->inicializa_processo();
  srand(M->id_processo());

  /*Inicializa  o contador de mortos */
  contador_vivos = M->tamanho_rede() - 1; /*Descontando o monitor*/
  mortos = (int *) calloc (M->tamanho_rede(), sizeof(int));
  for (i=0; i<contador_vivos + 1; i++) {
    mortos[i] = 0;
  }
  mortos[M->monitor()] = 1; /*Considerar o monitor como morto sempre 
				- ele nao entra nas contagens*/

  /*Define o proximo processo no anel de tokens
    isso eh sempre constante*/

  proximo_processo = ((M->id_processo() + 1) % M->tamanho_rede());
  if (proximo_processo == M->monitor()) { 
      proximo_processo = ((proximo_processo + 1) % M->tamanho_rede());
  }

  if (M->id_processo() == 1) {
    /*Primeiro processo - aguarda o desbloqueio e gera um token
      Como nao foi implementado nenhum algoritmo de eleicao de lider
      ou coisa parecida, vai ser arbitrario mesmo 
      Poderia ser um codigo fonte diferente, mas .. nao tem 
      muita necessidade*/
    M->aguarda_desbloqueio();

    token = 1;
    contador_passadas = 1;

    B = new buffer();

    B->adiciona_pacote((void *)&token, MSG_INT, 1, TOKEN_TAG);
    B->adiciona_pacote((void *)&contador_passadas, MSG_INT, 1, TOKEN_TAG);    
    M->envia_mensagem(B, proximo_processo, TOKEN_TAG);
  }

  do {
    M->recebe_mensagem(&B, MSG_ANY_SOURCE, MSG_ANY_TAG, &s);
    if (s.tag != DIE_TAG && !morte_iniciada) {
      if (s.tag == TOKEN_TAG) {
	/*Recolhendo informacoes da mensagem recebida*/
	B->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	contador_passadas = *((int *)vetor);
	free(vetor);
	B->remove_pacote(&vetor, &tipo_p, &tamanho_p, &tag_p);
	token = *((int *)vetor);      
	free(vetor);
	delete(B);
	/*Atualizando informacoes sobre os pacotes*/
	contador_passadas++;
   
	if (contador_passadas <= LIMITE_PASSADAS) { 
	  if (M->id_processo() == 1) {
	    M->tira_checkpoint_basico();      
	  }
	  /*Criando novo buffer para envio da proxima mensagem*/
	  B = new buffer();
	  B->adiciona_pacote((void *)&token, MSG_INT, 1, TOKEN_TAG);
	  B->adiciona_pacote((void *)&contador_passadas, MSG_INT, 1, TOKEN_TAG);    
	  /*Repassando o token*/
	  M->envia_mensagem(B,proximo_processo, TOKEN_TAG);      
	  /*Tirando um checkpoint basico */
	  token = 0;
	} else {
	  /*Caso contrario, enviando uma mensagem de morte para todos 
	    os processos e depois saindo do laco*/
	  conta_processos = M->tamanho_rede();
	  B = new buffer();
	  for (i=0; i < conta_processos; i++) {
	    if ((i != M->id_processo()) && (i != M->monitor())) {
	      M->envia_mensagem(B, i, DIE_TAG);
	    }
	  }
	  /*Considera a morte inicializada para esse caso 
	    (somente um processo passara por aqui, pois para tal, ele deve
	    estar com o token)*/
	  morte_iniciada = 1;
	  contador_vivos--;
	  mortos[M->id_processo()] = 1;
	}
	delete(B);
	if (!morte_iniciada) {
	  /*Se a morte nao foi iniciada por esse processo, enviar mensagens aleatórias*/
	  B = new buffer();
	  mensagens_aleatorias = rand() % 5;
	  for (i=0; i < mensagens_aleatorias; i++) {
	    destino_aleatorio = (rand() % M->tamanho_rede());
	    if ((destino_aleatorio != M->monitor()) && (destino_aleatorio != M->id_processo()))  {
	      M->envia_mensagem(B, destino_aleatorio, RANDOM_TAG);
	    }
	  }
	  delete(B);	  
	}
      } 
     
    } else {
      /*Se recebeu um contador de morte, inicia o processo de morte
        caso ele ja esteja iniciado, teste para verificar se ele chegou ao 
        final*/
      if (!morte_iniciada) {
	/*Se a morte nao foi inicializada, inicializar a morte*/
	morte_iniciada = 1;
	conta_processos = M->tamanho_rede();
	B = new buffer();
	for (i=0; i < conta_processos; i++) {
	  if ((i != M->id_processo()) && (i != M->monitor())) {
	    M->envia_mensagem(B, i, DIE_TAG);
	  }
	}
	/*Considera a morte inicializada para esse caso 
	  (somente um processo passara por aqui, pois para tal, ele deve
	  estar com o token)*/
	morte_iniciada = 1;
	contador_vivos--;
	mortos[M->id_processo()] = 1;
	contador_vivos--;
	mortos[s.source] = 1;
      } else {
	/*Simplesmente decrementar o contador de morte se uma mensagem de
	  de morte for recebida e a tabela ja nao tiver conhecimento dela. 
	  Caso contrário, desconsiderar a mensagem*/
	if ((s.tag == DIE_TAG) && (!mortos[s.source])) {
	  contador_vivos--;
	  mortos[s.source] = 1;
	}
      }
    }
  } while (contador_vivos > 0);
  M->finaliza_processo();
  return (0);
}
